# -*- coding: utf-8 -*-

"""

Add SAML support to your Python softwares using this library.

SAML Python toolkit let you build a SP (Service Provider)
over your Python application and connect it to any IdP (Identity Provider).

Supports:

* SSO and SLO (SP-Initiated and IdP-Initiated).
* Assertion and nameId encryption.
* Assertion signature.
* Message signature: AuthNRequest, LogoutRequest, LogoutResponses.
* Enable an Assertion Consumer Service endpoint.
* Enable a Single Logout Service endpoint.
* Publish the SP metadata (which can be signed).
"""
